/*
 *  OpFlowTools_fieldOp.h
 *  
 *
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */

#ifndef DEF_OFTOOLS_FO
#define DEF_OFTOOLS_FO

#include "CImg.h"
#include "coeff.h"
#include "interpol.h"
#include "derivee_spline.h"
#include "io.h"
/*========VARIOUS TOOLS===========*/

/**
 *  dperiod
 */
inline double dperiod(double ind, int max);

/**
 *  filtre_2D
 */
cimg_library::CImg<double>  Filtre_2D(const cimg_library::CImg<double>& b,
                                      const cimg_library::CImg<double>& filtre);
cimg_library::CImg<double>  Filtre_2D(const cimg_library::CImg<double>& b,
                                      const cimg_library::CImg<double>& filtre,
                                      bool periodic);
cimg_library::CImg<double>  Filtre1DSymSize5_Im2D_Adj(const cimg_library::CImg<double>& b,
                                        const cimg_library::CImg<double>& filtre);
/**
 * projectionUV
 */
void projectionUV(cimg_library::CImg<double>& UVproj,
                  const cimg_library::CImg<double>& UV,
                  int SplineDegree);

/**
 * maj_vecteurs
 */
//-/!\/!\/!\/!\/!\/!\/!\ BUG with rectangular fields !!!!!!!!
void maj_vecteurs(cimg_library::CImg<double> &X,
                  const cimg_library::CImg<double> & Xin,
                  bool divNul);
/**
 * isotropicPSD
 */
cimg_library::CImg<double> isotropicPSD(const cimg_library::CImg<double> &UV,
                                        double dx=1.,
                                        double dt=1.);
cimg_library::CImg<double> isotropicPSD_Scalar(const cimg_library::CImg<double> &Im, double blurStdev, double dx, double dt);

cimg_library::CImg<double> isotropicPSD(const cimg_library::CImg<double> &UV,double blurStdev=1.,
                                        double dx=1.,
                                        double dt=1.);

/**
 * curl
 */
cimg_library::CImg<double> curlFFT(const cimg_library::CImg<double> &UVin);

/**
 * div
 */
cimg_library::CImg<double> divFFT(const cimg_library::CImg<double> &UVin);

cimg_library::CImg<double> zeroPadSampling(const cimg_library::CImg<double> &in, int fac);
cimg_library::CImg<double> decimate(const cimg_library::CImg<double> &in, int fac);

/*==================SPLINES-BASED TOOLS=======================*/
cimg_library::CImg<double>  coefficientsToSamples(const cimg_library::CImg<double> &coeff,
                           int SplineDegree);
cimg_library::CImg<double>  SamplesToCubicSplineCoefAdjoint(const cimg_library::CImg<double> &samples,const cimg_library::CImg<double> &D);
cimg_library::CImg<double>  DforAdjointCubicSpline(int width);
void Recal_spline(cimg_library::CImg<double> &out,
                  const cimg_library::CImg<double> &in,
                  const cimg_library::CImg<double> &UV,
                  int SplineDegree,
                  int boundaryCondition=0);

void Recal_splineOld(cimg_library::CImg<double> &out,
                  const cimg_library::CImg<double> &in,
                  const cimg_library::CImg<double> &UV,
                  int SplineDegree);

void Proj_Dup_Spline(cimg_library::CImg<double> &out,
                     const cimg_library::CImg<double> &in,
                     int SplineDegree);

cimg_library::CImg<double> gradSpline_X(const cimg_library::CImg<double> &in,
                                        double sigma,
                                        bool periodic = false);

cimg_library::CImg<double> gradSpline_Y(const cimg_library::CImg<double> &in,
                                        double sigma,
                                        bool periodic = false);
cimg_library::CImg<double> displacedGradSpline_X(const cimg_library::CImg<double> &in,const cimg_library::CImg<double> &UV);
cimg_library::CImg<double> displacedGradSpline_Y(const cimg_library::CImg<double> &in,const cimg_library::CImg<double> &UV);
cimg_library::CImg<double> gradSpline_XX(const cimg_library::CImg<double> &in,
                                         double sigma);

cimg_library::CImg<double> gradSpline_YY(const cimg_library::CImg<double> &in,
                                         double sigma);

cimg_library::CImg<double> gradSpline_XY(const cimg_library::CImg<double> &in,
                                         double sigma);

cimg_library::CImg<double> grad_t(const cimg_library::CImg<double> &in,
                                  const cimg_library::CImg<double> &indt,
                                  double sigma);

cimg_library::CImg<double> curlSpline(const cimg_library::CImg<double> &UVin,
                                      double sigma);

cimg_library::CImg<double> divSpline(const cimg_library::CImg<double> &UVin,
                                     double sigma);

/*=======FINITE-DIFFERENCE-BASED TOOLS===========*/

cimg_library::CImg<double> gradFD1_X(const cimg_library::CImg<double> &in,
                                     bool periodic);

cimg_library::CImg<double> gradFD1_Y(const cimg_library::CImg<double> &in,
                                     bool periodic);

cimg_library::CImg<double> gradFD1_X_Adj(const cimg_library::CImg<double> &in,
                           bool periodic);
cimg_library::CImg<double> gradFD1_Y_Adj(const cimg_library::CImg<double> &in,
                           bool periodic);

cimg_library::CImg<double> gradFD2_X(const cimg_library::CImg<double> &in,
                                     double sigma,
                                     bool periodic);

cimg_library::CImg<double> gradFD2_Y(const cimg_library::CImg<double> &in,
                                     double sigma,
                                     bool periodic);
cimg_library::CImg<double> gradFD2_X(const cimg_library::CImg<double> &in,
                                     bool periodic);

cimg_library::CImg<double> gradFD2_Y(const cimg_library::CImg<double> &in,
                                     bool periodic);
cimg_library::CImg<double> gradFD4_X(const cimg_library::CImg<double> &in,
                                     double sigma,
                                     bool periodic);

cimg_library::CImg<double> gradFD4_Y(const cimg_library::CImg<double> &in,
                                     double sigma,
                                     bool periodic);


cimg_library::CImg<double> grad_5thOrder(const cimg_library::CImg<double> &in, bool X, bool Y);

 cimg_library::CImg<unsigned char> map2middlebury(const cimg_library::CImgList<double>& flow);
/*=================CONVERSION======================*/

inline void float2double(double *out, const float *in , int Nel){
  for (int k=0; k<Nel; k++) out[k] = (double)in[k];
};

inline void double2float(float *out, const double *in , int Nel){
  for (int k=0; k<Nel; k++) out[k] = (float)in[k];
};

/*obsolete*/
double * CImg2double( const cimg_library::CImg<double> &in);
/*obsolete*/
float * CImg2float( const cimg_library::CImg<double> &in);

/*obsolete*/
cimg_library::CImg<double> double2CImg(int dx,
                                       int dy,
                                       double *in);
/*obsolete*/
cimg_library::CImg<double> float2CImg(int dx,
                                      int dy,
                                      float *in);
/*obsolete*/
void Copy_d(double *im1,
            double *im2,
            int nb_col,
            int nb_line);
/*obsolete*/
void Copy_f(float *im1,
            float *im2,
            int nb_col,
            int nb_line);

#endif
