/*
 *  OpFlowTools_StructFun.h
 *  
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */

#ifndef DEF_OFTOOLS_SF
#define DEF_OFTOOLS_SF

#include "CImg.h"

/*=======================STRUCTURE FUNCTIONS=from=Mr Heas===========================*/

double  power_LeastSquareEstimator(const cimg_library::CImg<double> &S2,
                                   const cimg_library::CImg<double> &S2_,
                                   int neta,
                                   int lo);

cimg_library::CImg<double> StructureFunctionPthAbs_horiVertDir(const cimg_library::CImg<double> &UV,
                                                               int neta,
                                                               int lo,
                                                               const cimg_library::CImgList <double> &z_right,
                                                               const cimg_library::CImgList <double> &z_down,
                                                               const cimg_library::CImgList <double> &z_t_right,
                                                               const cimg_library::CImgList <double> &z_t_down,
                                                               double  p);

cimg_library::CImg<double> StructureFunctionPthTransAbs_horiVertDir(const cimg_library::CImg<double> &UV,
                                                                    int neta,
                                                                    int lo,
                                                                    double p);

cimg_library::CImg<double> StructureFunctionPthLongAbs_horiVertDir(const cimg_library::CImg<double> &UV,
                                                                   int neta,
                                                                   int lo,
                                                                   double p);

cimg_library::CImg<double> StructureFunctionPthAbs_crossDir(const cimg_library::CImg<double> &UV,
                                                            int neta,
                                                            int lo,
                                                            const cimg_library::CImgList <double> &z_rightDown,
                                                            const cimg_library::CImgList <double> &z_rightUp,
                                                            const cimg_library::CImgList <double> &z_t_rightDown,
                                                            const cimg_library::CImgList <double> &z_t_rightUp,
                                                            double p);

cimg_library::CImg<double> StructureFunctionPthTransAbs_crossDir(const cimg_library::CImg<double> &UV,
                                                                 int neta,
                                                                 int lo,
                                                                 double p);

cimg_library::CImg<double> StructureFunctionPthLongAbs_crossDir(const cimg_library::CImg<double> &UV,
                                                                int neta,
                                                                int lo,
                                                                double p);


/*!
    @function
    @abstract   This is a basic linear fit of (X,Y) data in the sense of least-square errors.
    @discussion There is no ponderation. Could be added...
    @param      data is an N(columns)x2(lines) image. First line is X data, second is corresponding Y data.
    @param      slope [out] is the estimated slope.
    @param      origin [out] the estimated origin.
    @result     Difference between estimated straight line and input points: (slope*X-origin) - Y.
*/

cimg_library::CImg<double> LeastSquareFit(const cimg_library::CImg<double> &data,
                                          double &slope,
                                          double &origin);

/*!
    @function
    @abstract   Compute structure function at given order, working on image width (x).
    @param      In is the input signal.
    @param      order is the chosen order
    @param      incMax is the increment max value (in pixel). If non precised (or 0), used value will be incDef = In.width()-1.
    @result     A 1D image containing values of the structure function, for increments from 1 to incMax (or incDef...).
*/
cimg_library::CImg<double> StructureFunctionWidth(const cimg_library::CImg<double> &In,
                                                  double order,
                                                  int incMax=0);

/*!
 @function
 @abstract   Similar to @link StructureFunctionWidth StructureFunctionWidth@/link, but works on image height (y).
 */
cimg_library::CImg<double> StructureFunctionHeight(const cimg_library::CImg<double> &In,
                                                   double order,
                                                   int incMax=0);

/*!
 @function
 @abstract    Computes structure function for a given velocity field at a given order.
 @discussion  Computed SF can be longitudinal or/and transversal for U or/and V. This is set up with ux, uy, vx, vy flags.
              For instance, "classical" longitudinal SF is computed with ux=true and vy=true.
 @param       UV the motion field
 @param       order the resquested order for SF
 @param       ux enable U longitudinal SF
 @param       uy enable U transversal SF
 @param       vx enable U longitudinal SF
 @param       vy enable U transversal SF
 @param      incMax is the increment max value (in pixel). If non precised (or 0), used value will be incDef = In.width()-1.
 @result     A 1D image containing values of the structure function, for increments from 1 to incMax (or incDef...).
 */
cimg_library::CImg<double> StructureFunctionFullUV(const cimg_library::CImg<double> &UV,
                                                       double order,
                                                       bool ux=true,
                                                       bool uy=false,
                                                       bool vx=false,
                                                       bool vy=false,
                                                       int incMax=0);
#endif
