/*!
 *  @header OpFlowTools_visu.h
 *  
 *
 *  @copyright Copyright 2014 INRIA. All rights reserved.
 *
 */


#ifndef DEF_OFTOOLS_VISU
#define DEF_OFTOOLS_VISU

#include "CImg.h"

/*!
    @abstract   Returns a 256-color palette (Look-Up Table) identical to Matlab's Jet.
    @result     The 1x256x1x3 color palette. 
*/
cimg_library::CImg<double> jet_LUT256();

/*!
 @abstract   Returns a bilinear gradient (r1,g1,b1)->(r2,g2,b2) 256-color palette (Look-Up Table).
 @result     The 1x256x1x3 color palette. 
 */
cimg_library::CImg<double> gradient2_LUT256(int r1,
                                            int g1,
                                            int b1,
                                            int r2,
                                            int g2,
                                            int b2);

/*!
 @abstract   Returns a trilinear gradient (r1,g1,b1)->(r2,g2,b2)->(r3,g3,b3) 256-color palette (Look-Up Table).
 @result     The 1x256x1x3 color palette. 
 */
cimg_library::CImg<double> gradient3_LUT256(int r1,
                                            int g1,
                                            int b1,
                                            int r2,
                                            int g2,
                                            int b2,
                                            int r3,
                                            int g3,
                                            int b3);

/*!
  @abstract   Applies hard light filter. Formula from Gimp documentation :).
  @param      Image the filtered image
  @param      Mask the mask used for filtering
  @result     The filter image.
 */
cimg_library::CImg<double> hardLight(cimg_library::CImg<double>Image,
                                     cimg_library::CImg<double>Mask);

/*!
  @abstract   Applies scalar Hue-Saturation color code to given 2D motion field.
  @param      UV input motion field
  @param      normValue optional value for saturation normalization. if nul or omitted, max(norm(UV)) is used.
  @result     RVB-converted image, ready for display/export.
 */
cimg_library::CImg<double> motionColorCode(const cimg_library::CImg<double> &UV,
                                           double normValue=0.);

#endif