/*
 *  OpFlowTools_fieldOp.cpp
 *  
 *
 *  Copyright 2014 INRIA. All rights reserved.
 *
 */
#include <iostream>
#include "OpFlowTools_fieldOp.h"


#define AD(n,m)  ((n < 0) ? (0) : ((n > m-1) ? (m-1) : (n)))
#define PERIOD(n,m)  ((n < 0) ? (m-1) : ((n > m-1) ? (0) : (n)))
#define ABS(n)   ((n > 0) ? (n) : -(n))

using namespace cimg_library;
using namespace std;

/*========VARIOUS TOOLS===========*/

inline double dperiod(double ind,
                      int max){
  int iind = (int)ind;
  double dind = ind - iind;
  iind = (iind<0)? ((max + iind%max)%max) : iind%max;
  return ((dind<0.)&&(iind==0))? max + dind : iind + dind;
};


CImg<double>  Filtre_2D(const CImg<double>& b,
                        const CImg<double>& filtre){
  
  /*Filter with neumann boundary condition*/
  int nb_ligF=filtre.height(), nb_colF=filtre.width();
  int nb_line=b.height(), nb_col=b.width();
  int n= nb_colF/2; if (nb_colF==1){n=0;}
  int m= nb_ligF/2; if (nb_ligF==1){m=0;}
  CImg<double> a(nb_col,nb_line);
  for (int i=0;i<nb_line;i++)
  {
    for (int j=0;j<nb_col;j++)
    {
	    a(j,i)=0.0;
      
	    for (int k=-m;k<m+1;k++)
        for (int l=-n;l<n+1;l++)
        {
          a(j,i)+= b[AD(j+l,nb_col)+AD(i+k,nb_line)*nb_col] * filtre[(k+m)*nb_colF+(l+n)];
          
        }
    }
  }
  return a;
};
CImg<double>  Filtre1DSymSize5_Im2D_Adj(const CImg<double>& b,
                        const CImg<double>& filtre){
    
    CImg<double> Imtmp;
    Imtmp.assign(Filtre_2D(b,filtre));
    CImg<double> ImSup;
    ImSup.assign(Imtmp.get_fill(0.));
    if (filtre.width()==5){
        cimg_forY(ImSup,y){
            ImSup(0,y)=b(1,y)*filtre(0);
            ImSup(1,y)=-1*b(0,y)*filtre(0);
            ImSup(ImSup.width()-1,y)=b(ImSup.width()-2,y)*filtre(0);
            ImSup(ImSup.width()-2,y)=-1*b(ImSup.width()-1,y)*filtre(0);
        }
        
    }
    else{
        if (filtre.height()==5){
            cimg_forX(ImSup,x){
                ImSup(x,0)=b(x,1)*filtre(0,0);
                ImSup(x,1)=-1*b(x,0)*filtre(0,0);
                ImSup(x,ImSup.height()-1)=b(x,ImSup.height()-2)*filtre(0,0);
                ImSup(x,ImSup.height()-2)=-1*b(x,ImSup.height()-1)*filtre(0,0);
            }
            
        }
        else{cout<<"Error : Adj filter not implemented !"<<endl;}}
    return Filtre_2D(b,filtre)+ImSup;
  };


CImg<double>  Filtre_2D(const CImg<double>& b,
                        const CImg<double>& filtre,
                        bool periodic){
  
  if (periodic){
    /*Filter with periodic boundary condition*/
    int nb_ligF=filtre.height(), nb_colF=filtre.width();
    int nb_line=b.height(), nb_col=b.height();
    int n= nb_colF/2; if (nb_colF==1){n=0;}
    int m= nb_ligF/2; if (nb_ligF==1){m=0;}
    CImg<double> a(nb_col,nb_line);
    
    for (int i=0;i<nb_line;i++){
      for (int j=0;j<nb_col;j++){
        a(j,i)=0.0;
        
        for (int k=-m;k<m+1;k++)
          for (int l=-n;l<n+1;l++)
          {
            a(j,i)+= b[PERIOD(j+l,nb_col)+PERIOD(i+k,nb_line)*nb_col] * filtre[(k+m)*nb_colF+(l+n)];
          }
      }
    }
    return a;
  }
  else{return Filtre_2D(b,filtre);}
};

void projectionUV(CImg<double>& UVproj,
                  const CImg<double>& UV,
                  int SplineDegree){
  
  CImg<double> temp;
  
  UVproj.assign(2*UV.width(),2*UV.height(),2,1);
  Proj_Dup_Spline(temp, UV.get_slice(0), SplineDegree);
  UVproj.get_shared_plane(0) = temp;
  Proj_Dup_Spline(temp, UV.get_slice(1), SplineDegree);
  UVproj.get_shared_plane(1) = temp;
};

//-/!\/!\/!\/!\/!\/!\/!\ BUG with rectangular fields !!!!!!!!
void maj_vecteurs(CImg<double> &X,
                  const CImg<double> & Xin,
                  bool divNul) {
  int nbc,nbl;
  nbl=(int)Xin.height();
  nbc=(int)Xin.width();  
  
  CImgList<double> Fdiv(2,nbc,nbl,1,1);
  CImgList<double> Fcurl(2,nbc,nbl,1,1);     
  CImgList<double> Fu(2,nbc,nbl,1,1);
  CImgList<double> Fv(2,nbc,nbl,1,1);     
  CImgList<double> fx(2,nbc,nbl,1,1);  
  CImgList<double> fy(2,nbc,nbl,1,1);  
  CImgList<double> u_fft(2,nbc,nbl,1,1);  
  CImgList<double> v_fft(2,nbc,nbl,1,1);     
  CImg<double> ak(nbc,1,1,1);
  CImg<double> ak2(nbc,1,1,1);  
  CImg<double> bk(nbl,1,1,1);
  CImg<double> bk2(nbl,1,1,1);   
  Fu=Xin.get_shared_plane(0).get_FFT(false); 
  Fv=Xin.get_shared_plane(1).get_FFT(false); 
  
  
  for (int j=0;j<nbc/2+1;j++)
  {
    ak(j)  = j;
    ak2(j) = j*j;
  }
  
  
  for (int j=nbc/2+1;j<nbc;j++)
  {
    int jj = nbc-j;
    ak(j)  = -ak(jj);
    ak2(j) = ak2(jj);
  }
  
  
  for (int j=0;j<nbl/2+1;j++)
  {
    bk(j)  = j;
    bk2(j) = j*j;
  }
  
  
  for (int j=nbl/2+1;j<nbl;j++)
  {
    int jj = nbl-j;
    bk(j)  = -bk(jj);
    bk2(j) = bk2(jj);
  }	 
  
  for (int j=0;j<nbl;j++)
    for (int k=0;k<nbc;k++)
    {
      if (!divNul){
        Fdiv[1](k,j) = (Fu[0](k,j)* ak(k) *2*M_PI/nbc +Fv[0](k,j)* bk(j) *2*M_PI/nbl );
        Fdiv[0](k,j) = -(Fu[1](k,j)* ak(k) *2*M_PI/nbc +Fv[1](k,j)* bk(j)  *2*M_PI/nbl);
      }
      else{ Fdiv[1](k,j)=0.;Fdiv[0](k,j)=0.;}
      Fcurl[1](k,j) = (Fv[0](k,j)* ak(k) *2*M_PI/nbc -Fu[0](k,j)* bk(j) *2*M_PI/nbl );
      Fcurl[0](k,j) = -(Fv[1](k,j)* ak(k) *2*M_PI/nbc -Fu[1](k,j)* bk(j) *2*M_PI/nbl );
    }
  
  for (int j=1;j<nbl;j++)
    for (int k=1;k<nbc;k++)
    {
      fx[0](k,j) = (Fdiv[1](k,j)* ak(k) -Fcurl[1](k,j)* bk(j) )/ (ak2(k)+bk2(j));
	    fx[1](k,j) = (Fcurl[0](k,j)* bk(j)- Fdiv[0](k,j)* ak(k) )/ (ak2(k)+bk2(j));
	    
	    
      fy[0](k,j) =  (Fcurl[1](k,j)* ak(k)+Fdiv[1](k,j)* bk(j)) / (ak2(k)+bk2(j));
	    fy[1](k,j) =  (0.0-Fdiv[0](k,j)* bk(j)-Fcurl[0](k,j)* ak(k)) / (ak2(k)+bk2(j));
    }
  
  for (int j=1;j<nbl;j++)
  {
    fx[0](0,j) = 0.0-bk(j)*Fcurl[1](0,j) / bk2(j);
    fx[1](0,j) = bk(j)*Fcurl[0](0,j) / bk2(j);
    fy[0](0,j) =  bk(j)*Fdiv[1](0,j) / bk2(j); 
    fy[1](0,j) = 0.0-  bk(j)*Fdiv[0](0,j) / bk2(j);         
  }
  
  for (int j=1;j<nbc;j++)
  {	 
    fy[0](j,0) = ak(j)*Fcurl[1](j,0) / ak2(j);   
    fy[1](j,0) = 0.0-ak(j)*Fcurl[0](j,0) / ak2(j); 
    
    fx[0](j,0) =  ak(j)*Fdiv[1](j,0) / ak2(j);   
    fx[1](j,0) =  0.0- ak(j)*Fdiv[0](j,0) / ak2(j);   	 
  }  
  fx[0](0,0)=0.0;
  fy[0](0,0)=0.0;      
  fx[1](0,0)=0.0;
  fy[1](0,0)=0.0;
  
  u_fft=fx.get_FFT(true);
  v_fft=fy.get_FFT(true);    
  
  X.get_shared_plane(0) = u_fft[0]*sqrt(nbl*nbc)/(2*M_PI);
  X.get_shared_plane(1) = v_fft[0]*sqrt(nbl*nbc)/(2*M_PI);
};

CImg<double> isotropicPSD(const CImg<double> &UV,  double dx, double dt){
    return isotropicPSD(UV,  0., dx,  dt);
    
}
CImg<double> isotropicPSD_Scalar(const CImg<double> &Im, double blurStdev, double dx, double dt){
    CImg<double> UV;UV.assign(Im.width(),Im.height(),2,1,0.);
    UV.get_shared_plane(0)=Im;
    UV.get_shared_plane(1)=Im;
    return isotropicPSD(UV,  blurStdev, dx,  dt);
}

CImg<double> isotropicPSD(const CImg<double> &UV, double blurStdev, double dx, double dt){
  
  //----UV should be SQUARE, of length N=2^P
  //--this is ASSUMED for now.
  
  CImgList<double> fft2U, fft2V;
  CImg<double> tempRes, Res, E;
  CImg<int> intK;
  int N=UV.width(), nWN;
  long int z;
  double factE;
  double factK;
  
  //-multiplicative factors
  factE = .5*(dx/dt)*(dx/dt)*1./(8.*cimg::PI*cimg::PI);
  factK = (2.*cimg::PI);
  if (dx!=1.) { factK /= N*dx; }
  //-above there is N instead of (N-1) because we assume periodicity
  
  //-here we get 2d fft
  fft2U = UV.get_shared_plane(0).get_FFT();
  fft2V = UV.get_shared_plane(1).get_FFT();    
  E.assign( fft2U(0).sqr()+ fft2U(1).sqr() + fft2V(0).sqr() + fft2V(1).sqr() );
  
  //-we create (integer) wavenumbers vectors
  intK.assign(N,1,1,1, 0);
  //-first half is positive WN, second half is negative
  for (int k=1; k<N/2; k++) {
    intK(k) = k;
    intK(N-k) = -k;
  }
  intK(N/2) = N/2;
  
  //-we allocate oversized vectors for results
  //-max k2 = 2*(N/2)^2
  //-first ligne is k^2, second line Ek value, third one the number of points corresponding
  tempRes.assign(2*(N/2)*(N/2)+1,3,1,1, 0.);
  
  //-then we go
  nWN = 0;
  cimg_forXY(UV,i,j){
    //-this is the (integer) squared radius corresponding to current point
    z = intK(i)*intK(i) + intK(j)*intK(j);
    //-if we never visited this squared-radius value before
    if (tempRes(z,2)==0.){
      nWN++;
      tempRes(z,0) = sqrt((double)z);
    }
    //-add contribution (0.5 factor will be added later trough factE)
    tempRes(z,1) +=  E(i,j);
    //-and add a point 
    tempRes(z,2)++;
  }
    
  //-we now we have nNW values of PSD that we want to return
  Res.assign(nWN,2,1,1, 0.);
  z = 0;
  //-we browse tempRes
  cimg_forX(tempRes,p){
    //-if this is a valid point, get values 
    if (tempRes(p,2)!=0.) {
      Res(z,0) = factK*tempRes(p,0);
      Res(z,1) = Res(z,0)*factE*tempRes(p,1);
      z++;
    }
  }
  if (blurStdev!=0){  Res.get_shared_line(1).blur(blurStdev);}
  return Res;
};


CImg<double> curlFFT(const CImg<double> &UVin){
  
  CImgList<double> Fu = UVin.get_shared_plane(0).get_FFT(false); 
  CImgList<double> Fv = UVin.get_shared_plane(1).get_FFT(false); 
  CImgList<double> Fcurl(2,UVin.width(),UVin.height(),1,1);
  CImg<double> ak(UVin.width(),1,1,1), bk(UVin.height(),1,1,1);
  double wh, ww, fh=2.*M_PI/UVin.height(), fw=2.*M_PI/UVin.width();
  
  //-compute frequencies
  for (int j=0;j<UVin.width()/2+1;j++) {ak(j) = j;}
  for (int j=UVin.width()/2+1;j<UVin.width();j++) {ak(j) = -ak(UVin.width()-j);}
  for (int j=0;j<UVin.height()/2+1;j++) {bk(j) = j;}
  for (int j=UVin.height()/2+1;j<UVin.height();j++) {bk(j) = -bk(UVin.height()-j);}	 
  
  //-FFT(curl) coefficients  
  cimg_forXY(UVin,x,y){
    ww = ak(x)*fw;
    wh = bk(y)*fh;
    Fcurl(0)(x,y) = -ww*Fv(1)(x,y) + wh*Fu(1)(x,y);
    Fcurl(1)(x,y) = ww*Fv(0)(x,y) - wh*Fu(0)(x,y);
  }
  return Fcurl.get_FFT(true)(0);
};

CImg<double> divFFT(const CImg<double> &UVin){
  
  CImgList<double> Fu = UVin.get_shared_plane(0).get_FFT(); 
  CImgList<double> Fv = UVin.get_shared_plane(1).get_FFT(); 
  CImgList<double> Fdiv(2,UVin.width(),UVin.height(),1,1);
  CImg<double> ak(UVin.width(),1,1,1), bk(UVin.height(),1,1,1);
  double wh, ww, fh=2.*M_PI/UVin.height(), fw=2.*M_PI/UVin.width();
  
  //-compute frequencies
  for (int j=0;j<UVin.width()/2+1;j++) {ak(j) = j;}
  for (int j=UVin.width()/2+1;j<UVin.width();j++) {ak(j) = -ak(UVin.width()-j);}
  for (int j=0;j<UVin.height()/2+1;j++) {bk(j) = j;}
  for (int j=UVin.height()/2+1;j<UVin.height();j++) {bk(j) = -bk(UVin.height()-j);}	 
  //-FFT(div) coefficients
  cimg_forXY(UVin,x,y){
    ww = ak(x)*fw;
    wh = bk(y)*fh;
    Fdiv(0)(x,y) = -(ww*Fu(1)(x,y) + wh*Fv(1)(x,y));
    Fdiv(1)(x,y) = ww*Fu(0)(x,y) + wh*Fv(0)(x,y);
  }
  return Fdiv.get_FFT(true)(0);
};


CImg<double> decimate(const CImg<double> &in, int fac){
    CImg<double> out((int)((double)in.width()/(double)fac), (int)((double)in.height()/(double)fac),in.depth(),in.spectrum(),0.);
    cimg_forXYZC(out,x,y,z,c){out(x,y,z,c)=in(fac*x,fac*y,z,c);}
    return out;
};
CImg<double> zeroPadSampling(const CImg<double> &in, int fac){
    CImg<double> out(in.width()*fac, in.height()*fac,in.depth(),in.spectrum(),0.);
    cimg_forXYZC(in,x,y,z,c){out(fac*x,fac*y,z,c)=in(x,y,z,c);}
    return out;
};



/*==================SPLINES-BASED TOOLS=======================*/

/*------NEW STUFF 07/06/11--------------------------------
 Spline library has been converted to double precision.
 Functions below 
 No everything has been tested, please report any bug
 ---------------------------------------------------------*/

/*-------WARNING------------------------------------------
 Sadly, the spline library uses float[] signals as input.
 So, we're back to float here: each time we use splines, we have
 - to convert CImg<double>.data() to a float[]
 - use splines
 - then convert back from float[] to CImg<double>.data().
 This is achieved with (inline, see .h) double2float() and float2double(),
 those 2 functions work directly on arrays. No need to pass/return CImg<> objects.
 ---------------------------------------------------------*/

/**
 * rewritten on 07/05/2010
 *  07/06/11: using double-precision splines
 */
void Recal_spline(CImg<double> &out,
                  const CImg<double> &in,
                  const CImg<double> &UV,
                  int SplineDegree,
                  int boundaryCondition){
  
  double pI,qI, pMax, qMax;
  //-initializa coeff with copy of in
  CImg<double> coeff(in, false);
  
  //-assign output
  out.assign(in,"xy");

  //-then replace fIn by its spline coeffs
  if (SplineDegree>2){bool Error = SamplesToCoefficients(coeff.data(), coeff.width(), coeff.height(), SplineDegree);
    if(Error){printf("probleme\n");}
  }
  //-maximum coordinates
  pMax = (double)(out.width()-1);
  qMax = (double)(out.height()-1);
  cimg_forXY(out,p,q){
    //-compute coordinates where to interpolate
    // according to chosen boundary conditions.
    switch(boundaryCondition){
      default:
      case 0:
        //-neumann-0
        //----------
        pI = (double)p + UV(p,q,0);
        qI = (double)q + UV(p,q,1);
        //-if those coordinates are outside range, use border pixel value
        if (pI<0.) { pI = 0.;}
        else if (pI>pMax) { pI = pMax;}
        if (qI<0.) { qI = 0.;}
        else if (qI>qMax) { qI = qMax;}
        break;
      case 1:
        //-periodic
        //---------
        //-Remark: slight problem when pI in ]pMax, pMax+1[
        // because spline lib doesn't know that pMax+1 <=> 0
        // so that it applies its own border condition
        pI = dperiod((double)p + UV(p,q,0),out.width());
        qI = dperiod((double)q + UV(p,q,1),out.height());
        break;
      case 2:
        //-libspline implements mirror conditions
        //--------------------------------------
        pI = (double)p + UV(p,q,0);
        qI = (double)q + UV(p,q,1);
        //-nothing special to do here, spline functions will manage.
        break;
    }
    //-finally get inteprolated value
    out(p,q) = InterpolatedValue(coeff.data(), coeff.width(), coeff.height(), pI, qI, SplineDegree);
  }
};
CImg<double>  coefficientsToSamples(const CImg<double> &coeff_, int SplineDegree){
    
    //-initializa coeff with copy of in
    CImg<double> coeff(coeff_, false);
    CImg<double> samples;
    samples.assign(coeff,false);
    cimg_forXY(samples,p,q){
        samples(p,q) = InterpolatedValue(coeff.data(), coeff.width(), coeff.height(), p, q, SplineDegree);
    }
    return samples;
};

CImg<double>  SamplesToCubicSplineCoefAdjoint(const CImg<double> &samples,const CImg<double> &D){
    CImg<double> in(samples, false);

    if (samples.width()<8||samples.height()<8){cout<< "Error: Adjoint cubic spline transform not defined for image of size smaller than 16 x 16 pixels"<<endl;}
    else{    
        cimg_forY(in,y){
   
            CImg<double> line(in.get_crop(0,y,in.width()-1,y),false);
            CImg<double> lineCopy(line,false);
            SamplesToCoefficients(line.data(),line.width(), 1, 3);
            for (int i=0;i<line.width();i++){
                for (int j=0;j<8;j++){
                    //left border correction
                    line(j)+=lineCopy(i)* D(i,j);
                    //right border correction
                    line(line.width()-1-j)+=lineCopy(i)* D(i,line.width()-1-j);
                }
            }
            cimg_forX(line,x) {in(x,y)=line(x);}
        }
        
        cimg_forX(in,x){
            CImg<double> line(in.get_crop(x,0,x,in.height()-1),false);
            CImg<double> lineCopy(line,false);
            SamplesToCoefficients(line.data(),1,line.height(), 3);
            for (int i=0;i<line.height();i++){
                for (int j=0;j<8;j++){
                    //up border correction
                    line(0,j)+=lineCopy(0,i)* D(i,j);
                    //bottom border correction
                    line(0,line.height()-1-j)+=lineCopy(0,i)* D(i,line.height()-1-j);
                 }
            }
            cimg_forY(line,y){in(x,y)=line(0,y);}

        }
        
    }
    return in;
    
};

CImg<double>  DforAdjointCubicSpline(int width){
    
    CImg<double> res(width,width,1,1,0.);
    CImg<double> element(width,1,1,1,0.);
    cimg_forX(res,x){
        element.fill(0.);
        element(x)=1;
        SamplesToCoefficients(element.data(),element.width(), 1, 3);
        cimg_forY(res,y){res(x,y)=element(y);}
    }
    res=res.get_transpose()-res;
    return res;
    
};


/**
 * rewritten on 07/05/2010 [TO BE FIXED]
 * =>remove many unecessary memory duplication (image<=>array).
 */
void Proj_Dup_Spline(CImg<double> &out,
                     const CImg<double> &in,
                     int SplineDegree){
  };

/**
 * rewritten on 07/05/2010
 * =>simpler code, no memory copy
 *  07/06/11: using double-precision splines 
 *--------------------------------
 */
CImg<double> gradSpline_X(const CImg<double> &in,
                          double sigma,
                          bool periodic){
  
  int SplineDegree=3, pLast, pMax;
  CImg<double> out(in,false), coeff(in,"xy");
    
  //-smooth
  if (sigma) out.blur(sigma,true);
  if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree)){exit(1);}
  //-compute grad, neuman-0 border condition (?? plutot mirror !)
  gradX(coeff.data(),out.data(),in.width(),in.height(), SplineDegree,0);
  //-add borders if periodic (FD order 4)
  if (periodic){
    pMax = in.width()-1;
    pLast = pMax -1;
    cimg_forY(out,q){
      out(0,q) = (in(pLast,q) - 8.*in(pMax,q) + 8.* in(1,q) - in(2,q))/12.;
      out(pMax,q) = (in(pLast-1,q) - 8.*in(pLast,q) + 8.* in(0,q) - in(1,q))/12.;  
    }
  }
  return out;
};

/**
 * rewritten on 07/05/2010
 * =>simpler code, no memory copy
 *  07/06/11: using double-precision splines 
 *--------------------------------
 */
CImg<double> gradSpline_Y(const CImg<double> &in,
                          double sigma,
                          bool periodic){
  
  int SplineDegree=3, qMax, qLast;
  CImg<double> out(in,false), coeff(in,"xy");
  
  //-smooth
  if (sigma){out.blur(sigma,true);}
  
  //-get spline coeff
  if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree )){exit(1);}
  //-compute grad, neuman-0 border condition
  gradY(coeff.data(),out.data(),out.width(),out.height(), SplineDegree,0);
  //-add borders if periodic
  if (periodic){
    qMax = in.height()-1;
    qLast = qMax-1;
    cimg_forX(out,p){
      out(p,0) = (in(p,qLast) - 8.*in(p,qMax) + 8.*in(p,1) - in(p,2))/12.;
      out(p,qMax) = (in(p,qLast-1) - 8.*in(p,qLast) + 8.*in(p,0) - in(p,1))/12.;      
    }
  }
  //-end
  return out;
};
/**
 * written on 2014
 *--------------------------------
 */
CImg<double> displacedGradSpline_X(const CImg<double> &in,const CImg<double> &UV){
    
    int SplineDegree=3, pLast, pMax;
    CImg<double> out(in,false), coeff(in,"xy");
    CImg<double> displacement(UV,false);
    //-smooth
    if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree)){exit(1);}
    //-compute grad displaced
    coefCubicDisplacedGradX(coeff.data(),displacement.data(),out.data(),in.width(),in.height());
    //-add borders if periodic (FD order 4)
    return out;
};

/**
 * written on 2014
*--------------------------------
 */
CImg<double> displacedGradSpline_Y(const CImg<double> &in,const CImg<double> &UV){
    
    int SplineDegree=3, pLast, pMax;
    CImg<double> out(in,false), coeff(in,"xy");
    CImg<double> displacement(UV,false);
    //-smooth
    if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree)){exit(1);}
    //-compute grad displaced
    coefCubicDisplacedGradY(coeff.data(),displacement.data(),out.data(),in.width(),in.height());
    //-add borders if periodic (FD order 4)
    return out;
    
 };



/**
 * rewritten on 07/05/2010
 * =>simpler code, no memory copy
 *  07/06/11: using double-precision splines 
 */
CImg<double> gradSpline_XX(const CImg<double> &in,
                           double sigma){
  
  int SplineDegree=3;
  CImg<double> out(in,false), coeff(in,"xy");
  
  //-smooth
  if (sigma){out.blur(sigma,true);}
  //-spline coeffs
  if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree )){exit(1);}
  //-grad
  derivXX(coeff.data(),out.data(),out.width(),out.height(), SplineDegree,0);
  return out;  
};

/**
 * rewritten on 07/05/2010
 * =>simpler code, no memory copy
 *  07/06/11: using double-precision splines 
 */
CImg<double> gradSpline_YY(const CImg<double> &in,
                           double sigma){
  
  int SplineDegree=3;
  CImg<double> out(in,false), coeff(in,"xy");
  
  //-smooth
  if (sigma){out.blur(sigma,true);}
  //-spline coeffs
  if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree )){exit(1);}
  derivYY(coeff.data(),out.data(),out.width(),out.height(), SplineDegree,0);
  return out;  
};

/**
 * rewritten on 07/05/2010
 * =>simpler code, no memory copy
 *  07/06/11: using double-precision splines 
 */
CImg<double> gradSpline_XY(const CImg<double> &in,
                           double sigma){
  
  int SplineDegree=3;
  CImg<double> out(in,false), coeff(in,"xy");
  
  //-smooth
  if (sigma){out.blur(sigma,true);}
  //-spline coeffs
  if (directBsplineMirror(out.data(), coeff.data(), out.width(), out.height(), 1,SplineDegree )){exit(1);}
  derivXY(coeff.data(),out.data(),out.width(),out.height(), SplineDegree,0);
  return out;  
  
};

/**
 * rewritten on 07/05/2010
 * =>simpler code
 */
CImg<double> grad_t(const CImg<double> &in,
                    const CImg<double> &indt,
                    double sigma){
	
  CImg<double> out(indt-in);
  
  if (sigma)out.blur(sigma,true);
  return out;	
};

/**
 * written on 08/06/2010
 */
CImg<double> curlSpline(const CImg<double> &UVin,
                        double sigma){
  return gradSpline_X(UVin.get_shared_plane(1), sigma) - gradSpline_Y(UVin.get_shared_plane(0), sigma);
}

/**
 * written on 08/06/2010
 */
CImg<double> divSpline(const CImg<double> &UVin,
                       double sigma){
  return gradSpline_X(UVin.get_shared_plane(0), sigma) + gradSpline_Y(UVin.get_shared_plane(1), sigma);
}

/*=======FINITE-DIFFERENCE-BASED TOOLS===========*/

CImg<double> gradFD1_X(const CImg<double> &in,
                      bool periodic){
  
    CImg<double> filtre(3);
    filtre(0)=-1.;filtre(1)=1.;filtre(2)=0.;
   return Filtre_2D(in,filtre,periodic);
    
};
CImg<double> gradFD1_X_Adj(const CImg<double> &in,
                       bool periodic){
    CImg<double> tmp;tmp.assign(in,"xyzc",0.);
    CImg<double> filtre(3);
    filtre(0)=0.;filtre(1)=1.;filtre(2)=-1.;
    tmp=Filtre_2D(in,filtre,periodic);
    cimg_forY(tmp,y){tmp(0,y)=-1*in(1,y);tmp(tmp.width()-1,y)=1*in(tmp.width()-1,y);}
    return tmp;
    
};
CImg<double> gradFD1_Y_Adj(const CImg<double> &in,
                           bool periodic){
    CImg<double> tmp;tmp.assign(in,"xyzc",0.);
    CImg<double> filtre(1,3);
    filtre(0,0)=0.;filtre(0,1)=1.;filtre(0,2)=-1.;
    tmp=Filtre_2D(in,filtre,periodic);
    cimg_forX(tmp,x){tmp(x,0)=-1*in(x,1);tmp(x,tmp.height()-1)=1*in(x,tmp.height()-1);}
    return tmp;
    
};


CImg<double> gradFD1_Y(const CImg<double> &in,
                      bool periodic){
  
    CImg<double> filtre(1,3);
    filtre(0,0)=-1.;filtre(0,1)=1.;filtre(0,2)=0.;
    return Filtre_2D(in,filtre,periodic);
};

CImg<double> gradFD2_X(const CImg<double> &in,
                       double sigma,
                       bool periodic){
    
    //-initialize gradient
    CImg<double> sIn(in,false);
    //-smooth
    if (sigma){ sIn.blur(sigma,true); }
    
    CImg<double> filtre(3);
    filtre(0)=-0.5;filtre(1)=0.;filtre(2)=0.5;
    return Filtre_2D(sIn,filtre,periodic);

};

CImg<double> gradFD2_Y(const CImg<double> &in,
                       double sigma,
                       bool periodic){
    
    //-initialize gradient
    CImg<double> sIn(in,false);
    //-smooth
    if (sigma){ sIn.blur(sigma,true); }
    
    CImg<double> filtre(1,3);
    filtre(0,0)=-0.5;filtre(0,1)=0.;filtre(0,2)=0.5;
    return Filtre_2D(sIn,filtre,periodic);

};
CImg<double> gradFD2_X(const CImg<double> &in,
                       bool periodic){
    
    CImg<double> filtre(3);
    filtre(0)=-0.5;filtre(1)=0.;filtre(2)=0.5;
    return Filtre_2D(in,filtre,periodic);
    
};

CImg<double> gradFD2_Y(const CImg<double> &in,
                       bool periodic){
     
    CImg<double> filtre(1,3);
    filtre(0,0)=-0.5;filtre(0,1)=0.;filtre(0,2)=0.5;
    return Filtre_2D(in,filtre,periodic);
    
};

CImg<double> gradFD4_X(const CImg<double> &in,
                       double sigma,
                       bool periodic){
  
  //-initialize gradient
  CImg<double> out(in,"xy"), sIn(in,false);
  int pMax = out.width()-1, pLast = out.width()-2;
  
  //-smooth
  if (sigma) sIn.blur(sigma,true);
  
  //-loop on all lines
  cimg_forY(out,q){
    //-loop on inner values
    //---------------------
    for (int p=2; p<pLast; p++) {
      out(p,q) = (sIn(p-2,q) - 8.*sIn(p-1,q) + 8.* sIn(p+1,q) - sIn(p+2,q))/12.;
    }
    //-then vertical boundary region
    //------------------------------
    //-periodic case, order 4 everywhere
    if (periodic) {
      out(0,q) = (sIn(pLast,q) - 8.*sIn(pMax,q) + 8.* sIn(1,q) - sIn(2,q))/12.;
      out(1,q) = (sIn(pMax,q) - 8.*sIn(0,q) + 8.* sIn(2,q) - sIn(3,q))/12.;
      out(pLast,q) = (sIn(pLast-2,q) - 8.*sIn(pLast-1,q) + 8.* sIn(pMax,q) - sIn(0,q))/12.;
      out(pMax,q) = (sIn(pLast-1,q) - 8.*sIn(pLast,q) + 8.* sIn(0,q) - sIn(1,q))/12.;      
    }
    //-neumann-0 case, we consider order 2 on the border to keep order 4 on border+-1
    else{
      out(0,q) = 0.;
      out(1,q) = (-8.*sIn(0,q) + sIn(1,q) + 8.* sIn(2,q) - sIn(3,q))/12.; 
      out(pLast,q) = (sIn(pLast-2,q) - 8.*sIn(pLast-1,q) - sIn(pLast,q) + 8.* sIn(pMax,q))/12.;
      out(pMax,q) = 0.; 
    }
  }
  return out;
};

CImg<double> gradFD4_Y(const CImg<double> &in,
                       double sigma,
                       bool periodic){
  //-initialize gradient
  CImg<double> out(in,"xy"), sIn(in,false);
  int qMax = out.height()-1, qLast = out.height()-2;
  
  //-smooth
  if (sigma) sIn.blur(sigma,true);
  
  //-loop on all columns
  cimg_forY(out,p){
    //-loop on inner values
    for (int q=2; q<qLast; q++) {
      out(p,q) = (sIn(p,q-2) - 8.*sIn(p,q-1) + 8.*sIn(p,q+1) - sIn(p,q+2))/12.;
    }
    //-then horizontal boundary region
    //------------------------------
    //-periodic case, order 4 everywhere
    if (periodic) {
      out(p,0) = (sIn(p,qLast) - 8.*sIn(p,qMax) + 8.*sIn(p,1) - sIn(p,2))/12.;
      out(p,1) = (sIn(p,qMax) - 8.*sIn(p,0) + 8.*sIn(p,2) - sIn(p,3))/12.;      
      out(p,qLast) = (sIn(p,qLast-2) - 8.*sIn(p,qLast-1) + 8.*sIn(p,qMax) - sIn(p,0))/12.;
      out(p,qMax) = (sIn(p,qLast-1) - 8.*sIn(p,qLast) + 8.*sIn(p,0) - sIn(p,1))/12.;      
    }
    //-neumann-0 case, we consider order 2 on the border to keep order 4 on border+-1
    else{
      out(p,0) = 0.;
      out(p,1) = (-8.*sIn(p,0) + sIn(p,1) + 8.*sIn(p,2) - sIn(p,3))/12.;      
      out(p,qLast) = (sIn(p,qLast-2) - 8.*sIn(p,qLast-1) - sIn(p,qLast) + 8.*sIn(p,qMax))/12.;
      out(p,qMax) = 0.;
    }
  }
  return out;
}


//=================
//------------methods 5th order finite difference shcemes
//=================

CImg<double> grad_5thOrder(const CImg<double> &in, bool X, bool Y){
    
    CImg<double> tmp, res(in,0.);
    tmp.assign(in);
    int Lx=res.width(),Ly=res.height();
    double dx=1./Lx;
    double a=3./2.,b=-1.*3./5.,c=1./10.;
    if (X){
        cimg_forXY(in,x,y){
            res(x,y)   = a/(2.*dx)*(tmp(dperiod(x+1,Lx),y)-tmp(dperiod(x-1,Lx),y))
            +b/(4.*dx)*(tmp(dperiod(x+2,Lx),y)-tmp(dperiod(x-2,Lx),y))
            +c/(6.*dx)*(tmp(dperiod(x+3,Lx),y)-tmp(dperiod(x-3,Lx),y));
        }
        tmp=res;
    }
    if (Y){
        cimg_forXY(in,x,y){
            res(x,y)   = a/(2.*dx)*(tmp(x,dperiod(y+1,Ly))-tmp(x,dperiod(y-1,Ly)))
            +b/(4.*dx)*(tmp(x,dperiod(y+2,Ly))-tmp(x,dperiod(y-2,Ly)))
            +c/(6.*dx)*(tmp(x,dperiod(y+3,Ly))-tmp(x,dperiod(y-3,Ly)));
        }
    }
    return res;
};




    
    /////////////////
    // MIDDLEBURY //
    ///////////////
    
    /**
     * The Middlebury color wheel.
     */
    static const CImg<unsigned char> middlebury_wheel(55, 1, 1, 3,
                                                                    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,	// RY (red)
                                                                    255, 212, 170, 127, 85, 42,													// YG (red)
                                                                    0, 0, 0, 0, 																// GC (red)
                                                                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,											// CB (red)
                                                                    0, 19, 39, 58, 78, 98, 117, 137, 156, 176, 196, 215, 235,					// BM (red)
                                                                    255, 255, 255, 255, 255, 255,												// MR (red)
                                                                    0, 17, 34, 51, 68, 85, 102, 119, 136, 153, 170, 187, 204, 221, 238,			// RY (green)
                                                                    255, 255, 255, 255, 255, 255,												// YG (green)
                                                                    255, 255, 255, 255,															// GC (green)
                                                                    255, 231, 208, 185, 162, 139, 115, 92, 69, 46, 23,							// CB (green)
                                                                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,										// BM (green)
                                                                    0, 0, 0, 0, 0, 0,															// MR (green)
                                                                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 								// RY (blue)
                                                                    0, 0, 0, 0, 0, 0,															// YG (blue)
                                                                    0, 63, 127, 191,															// GC (blue)
                                                                    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,						// CB (blue)
                                                                    255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,			// BM (blue)
                                                                    255, 212, 170, 127, 85, 42);												// MR (blue)
    
    
    /**
     * Convert a 2-dimensional field into a color one using the Middlebury color code.
     * \param flow the x- and y-coordinates of the field
     * \param linear_func if true, the value represented in white is 0 (minimum velocity otherwise)
     */

CImg<unsigned char> map2middlebury(const CImgList<double>& flow) {
        bool linear_func=true;
        
        const CImg<double>& vx = flow[0];
        const CImg<double>& vy = flow[1];
        
        CImg<unsigned char> colored(vx, "xyz3", 0);
        const int colors_nb = middlebury_wheel.width();
        
        /* velocity norm */
        CImg<float> rad = vx.get_sqr() + vy.get_sqr();
        rad.sqrt();
        if(linear_func)
            rad /= rad.max();
        else
            rad.normalize(0.F, 1.F);
        
        /* velocity angle */
        CImg<float> angle = -vy;
        angle.atan2(-vx);
        angle /= cimg::PI;
        
        /* colorize */
//#pragma omp parallel for collapse(3)
        cimg_forXYZ(colored, x, y, z) {
            
            float fk = (angle(x, y, z)+1.0)/2.0*(colors_nb-1);
            int k0 = (int)fk;
            int k1 = (k0+1)%colors_nb;
            float f = fk-k0;
            
            cimg_forC(colored, c) {
                float col0 = middlebury_wheel(k0, 0, 0, c);
                float col1 = middlebury_wheel(k1, 0, 0, c);
                float col = (1-f)*col0 + f*col1;
                if (rad(x, y, z) <= 1)
                    col = 255.0-rad(x, y, z)*(255.0-col); // increase saturation with radius
                else
                    col *= .75; // out of range
                colored(x, y, z, c) = col;
            }
        }
        
        return colored;
    }
    
 


/*=================OBSOLETE======================*/

/*--WARNING-----------------------------
 The 4 funtions below should not be used.
 CImg<> implementation allows to work directly on the data array,
 through .data() method to access the pointer.
 They are left here for (hypothetic) backward-compatibility :).
 --------------------------------------*/

/*obsolete*/
double * CImg2double( const CImg<double> &in){
  int dx=in.width();
  int dy=in.height();
  double *out=new double[dx*dy];
  for (int i=0;i<dy;i++) {
    for (int j=0;j<dx;j++){
      out[i*dx+j]= (double) in(j,i);
    }
  }
  return out;
};

/*obsolete*/
CImg<double> double2CImg(int dx, int dy, double *in){
  CImg<double> out(dx,dy);
  for (int i=0;i<dy;i++) {
    for (int j=0;j<dx;j++){
      out(j,i)=(double) in[i*dx+j];
    }
  }
  return out;
};

/*obsolete*/
float * CImg2float( const CImg<double> &in){
  int dx=in.width();
  int dy=in.height();
  float *out=new float[dx*dy];
  for (int i=0;i<dy;i++) {
    for (int j=0;j<dx;j++){
      out[i*dx+j]= (float) in(j,i);
    }
  }
  return out;
};

/*obsolete*/
CImg<double> float2CImg(int dx,
                        int dy,
                        float *in){
  CImg<double> out(dx,dy);
  for (int i=0;i<dy;i++) {
    for (int j=0;j<dx;j++){
      out(j,i)=(double) in[i*dx+j];
    }
  }
  return out;
};

/*--WARNING-----------------------------
 The 2 functions below only mimic memcpy(), 
 which for sure is better implemented.
 Also left here for bacward compatibility.
--------------------------------------*/

/*obsolete*/
void Copy_f(float *im1,
            float *im2,
            int nb_col,
            int nb_line){
  long k;
  register int i;
  k=(nb_line*nb_col);
  for (i=0;i<k;i++,im1++,im2++)
  {
    *im1= *im2;
  }
};

/*obsolete*/
void Copy_d(double *im1,
            double *im2,
            int nb_col,
            int nb_line){
  long k;
  register int i;
  k=(nb_line*nb_col);
  for (i=0;i<k;i++,im1++,im2++)
  {
    *im1= *im2;
  }
};
